# -*- coding: utf-8 -*-

import json
import time
import urllib
import logging

from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, BaseSocketHandler, auth_check
from litepipeline.manager.models.workflows import Workflows
from litepipeline.manager.models.works import Works
from litepipeline.manager.utils.scheduler import Scheduler
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, Errors, Stage, Status, Signal, splitall, JSONLoadError
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class CreateWorkHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            name = self.get_json_argument("name", "")
            workflow_id = self.get_json_argument("workflow_id", "")
            input_data = self.get_json_argument("input_data", {})
            workflow = Workflows.instance().get(workflow_id)
            if workflow and name:
                if not isinstance(input_data, dict):
                    raise JSONLoadError("input_data must be dict type")
                if workflow["enable"]:
                    work_id = Works.instance().add(name, workflow_id, input_data = input_data, configuration = workflow["configuration"])
                    if work_id is not False:
                        result["work_id"] = work_id
                    else:
                        Errors.set_result_error("OperationFailed", result)
                else:
                    Errors.set_result_error("WorkflowDisabled", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("CreateWorkHandler, name: %s, workflow_id: %s", name, workflow_id)
        except JSONLoadError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class RunWorkHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            work_id = self.get_json_argument("work_id", "")
            if work_id:
                work = Works.instance().get(work_id)
                if work:
                    workflow = Workflows.instance().get(work["workflow_id"])
                    if workflow:
                        if workflow["enable"]:
                            if work["stage"] == Stage.finished:
                                success = Works.instance().update(work_id, {"stage": Stage.pending, "status": None})
                                if not success:
                                    Errors.set_result_error("OperationFailed", result)
                            else:
                                Errors.set_result_error("WorkStillRunning", result)
                        else:
                            Errors.set_result_error("WorkflowDisabled", result)
                    elif workflow is None:
                        Errors.set_result_error("WorkflowNotExists", result)
                    else:
                        Errors.set_result_error("OperationFailed", result)
                elif work is None:
                    Errors.set_result_error("WorkNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("RunWorkHandler, work_id: %s", work_id)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class RecoverWorkHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            work_id = self.get_json_argument("work_id", "")
            if work_id:
                work = Works.instance().get(work_id)
                if work:
                    workflow = Workflows.instance().get(work["workflow_id"])
                    if workflow:
                        if workflow["enable"]:
                            if work["stage"] == Stage.finished:
                                if work["status"] != Status.success:
                                    success = Works.instance().update(work_id, {"stage": Stage.recovering, "status": None})
                                    if not success:
                                        Errors.set_result_error("OperationFailed", result)
                                else:
                                    Errors.set_result_error("WorkAlreadySuccess", result)
                            else:
                                Errors.set_result_error("WorkStillRunning", result)
                        else:
                            Errors.set_result_error("WorkflowDisabled", result)
                    elif workflow is None:
                        Errors.set_result_error("WorkflowNotExists", result)
                    else:
                        Errors.set_result_error("OperationFailed", result)
                elif work is None:
                    Errors.set_result_error("WorkNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("RecoverWorkHandler, work_id: %s", work_id)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class StopWorkHandler(BaseHandler): # kill -9 or -15
    @auth_check
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            work_id = self.get_json_argument("work_id", "")
            signal = int(self.get_json_argument("signal", Signal.kill))
            if work_id and signal in (Signal.kill, Signal.terminate):
                work = Works.instance().get(work_id)
                if work:
                    Works.instance().update(work_id, {"stage": Stage.stopping})
                    if work["stage"] in (Stage.pending, Stage.recovering):
                        Works.instance().update(work_id, {"stage": Stage.finished, "status": Status.kill})
                    elif work["stage"] == Stage.running:
                        for app_name in work["result"]:
                            app = work["result"][app_name]
                            if "stage" not in app or ("stage" in app and app["stage"] != Stage.finished):
                                success = yield Scheduler.instance().stop_task(app["task_id"], signal)
                                if not success:
                                    Errors.set_result_error("OperationFailed", result)
                                    break
                    elif work["stage"] == Stage.finished:
                        Errors.set_result_error("WorkAlreadyFinished", result)
                elif Work is None:
                    Errors.set_result_error("WorkNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("StopWorkHandler, work_id: %s, signal: %s", work_id, signal)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteWorkHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            work_id = self.get_argument("work_id", "")
            if work_id:
                success = Works.instance().delete(work_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class InfoWorkHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            work_id = self.get_argument("work_id", "")
            if work_id:
                work_info = Works.instance().get(work_id)
                if work_info:
                    result["info"] = work_info
                elif work_info is None:
                    Errors.set_result_error("WorkNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ListWorkHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            filters = {}
            work_id = self.get_argument("work_id", "")
            if work_id:
                filters["work_id"] = work_id
            workflow_id = self.get_argument("workflow_id", "")
            if workflow_id:
                filters["workflow_id"] = workflow_id
            name = self.get_argument("name", "")
            if name:
                filters["name"] = name
            stage = self.get_argument("stage", "")
            if stage:
                filters["stage"] = stage
            status = self.get_argument("status", "")
            if status:
                filters["status"] = status
            LOG.debug("ListWorkHandler offset: %s, limit: %s, filters: %s", offset, limit, filters)
            r = Works.instance().list(offset = offset, limit = limit, filters = filters)
            result["works"] = r["works"]
            result["total"] = r["total"]
            result["offset"] = offset
            result["limit"] = limit
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class CleanWorkHistoryHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {}
        self.write(result)
        self.finish()



# -*- coding: utf-8 -*-

import os
import time
import json
import hashlib
import logging
import datetime
from base64 import b64decode, b64encode

from tornado import ioloop
from tornado import gen
from tea_encrypt import EncryptStr, DecryptStr

from litepipeline.viewer.config import CONFIG

LOG = logging.getLogger(__name__)

BUF_SIZE = 65536


class Servers(object):
    HTTP_SERVER = None
    DB_SERVERS = []
    TORNADO_INSTANCE = None


async def shutdown():
    LOG.info("Stopping Service(%s:%s)", CONFIG["http_host"], CONFIG["http_port"])
    if Servers.HTTP_SERVER:
        Servers.HTTP_SERVER.stop()
        LOG.info("Stop http server!")
    for db_server in Servers.DB_SERVERS:
        db_server.close()
        LOG.info("Stop db server!")
    await gen.sleep(1)
    LOG.info("Will shutdown ...")
    ioloop.IOLoop.current().stop()


def sig_handler(sig, frame):
    LOG.warning("sig_handler Caught signal: %s", sig)
    ioloop.IOLoop.current().add_callback_from_signal(shutdown)


class Errors(object):
    OK = "ok"
    errors = {
        "ServerException": {"name": "ServerException", "message": "server exception"},
        "InvalidParameters": {"name": "InvalidParameters", "message": "invalid parameters"},
        "OperationFailed": {"name": "OperationFailed", "message": "operation failed"},
        "AppNotExists": {"name": "AppNotExists", "message": "application not exists"},
        "TaskNotExists": {"name": "TaskNotExists", "message": "task not exists"},
        "OperationRunning": {"name": "OperationRunning", "message": "operation running"},
        "WorkspaceNotExists": {"name": "WorkspaceNotExists", "message": "workspace not exists"},
        "WorkspaceNotPacked": {"name": "WorkspaceNotPacked", "message": "workspace not packed"},
    }

    @classmethod
    def set_result_error(cls, error_name, result, message = ""):
        if error_name in cls.errors:
            result["result"] = error_name
            if message == "":
                result["message"] = cls.errors[error_name]["message"]
            else:
                result["message"] = message
        else:
            result["result"] = "UnknownError"
            result["message"] = "unknown error"


class Stage(object):
    pending = "pending"
    running = "running"
    finished = "finished"


class Status(object):
    fail = "fail"
    success = "success"
    kill = "kill"
    cancel = "cancel"
    terminate = "terminate"


class Signal(object):
    kill = -9
    terminate = -15
    stop = -15
    cancel = -50


class JSONLoadError(Exception):
    def __init__(self, message):
        self.message = message


class MetaNotDictError(Exception):
    def __init__(self, message):
        self.message = message


def file_sha1sum(file_path):
    sha1 = hashlib.sha1()
    with open(file_path, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()


def file_md5sum(file_path):
    md5 = hashlib.md5()
    with open(file_path, 'rb') as fp:
        while True:
            data = fp.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


def bytes_md5sum(b):
    md5 = hashlib.md5()
    md5.update(b)
    return md5.hexdigest()


def encode_token(user, password):
    return b64encode(EncryptStr(user.encode("utf-8"),
                                bytes_md5sum(password.encode("utf-8"))
    ))


def splitall(path):
    allparts = []
    while True:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts


def get_workspace_path(create_at, task_id = None, action_name = None):
    result = ""
    date_create_at = datetime.datetime.strptime(create_at, "%Y-%m-%d %H:%M:%S.%f")
    date_directory_name = date_create_at.strftime("%Y-%m-%d")
    if task_id and action_name:
        result = os.path.join(CONFIG["data_path"], "tmp", "workspace", date_directory_name, task_id[:2], task_id[2:4], task_id, action_name)
    elif task_id:
        result = os.path.join(CONFIG["data_path"], "tmp", "workspace", date_directory_name, task_id[:2], task_id[2:4], task_id)
    else:
        result = os.path.join(CONFIG["data_path"], "tmp", "workspace", date_directory_name)
    return result


def init_storage():
    directories = [
        os.path.join(CONFIG["data_path"], "applications"),
        os.path.join(CONFIG["data_path"], "tmp", "workspace"),
        os.path.join(CONFIG["data_path"], "tmp", "download"),
    ]
    for d in directories:
        if not os.path.exists(d) or not os.path.isdir(d):
            os.makedirs(d)



# -*- coding: utf-8 -*-

import os
import sys
import time
import json
import logging
import shutil
import tarfile
import zipfile
from zipfile import ZipFile, ZipInfo
import signal
from pathlib import Path
import threading
from threading import Thread, Lock
from multiprocessing import Process, Queue, Pipe
import configparser

from tornado import gen, locks
import requests
from litepipeline_helper.models.client import LitePipelineClient, OperationFailedError

from litepipeline.node.utils.common import file_sha1sum, splitall, update_venv_cfg, StoppableThread, ZipFileWithPermissions, is_uuid
from litepipeline.node.config import CONFIG
from litepipeline import logger

LOG = logging.getLogger(__name__)


class Command(object):
    check_app = "check_app"
    exit = "exit"


class TasksCache(object):
    tasks = {}
    tasks_lock = Lock()

    @classmethod
    def set(cls, app_id):
        cls.tasks_lock.acquire()
        if app_id not in cls.tasks:
            cls.tasks[app_id] = False
        cls.tasks_lock.release()

    @classmethod
    def get(cls):
        result = False
        cls.tasks_lock.acquire()
        try:
            for app_id in cls.tasks:
                if not cls.tasks[app_id]:
                    cls.tasks[app_id] = True
                    result = app_id
                    break
        except Exception as e:
            LOG.exception(e)
        cls.tasks_lock.release()
        return result

    @classmethod
    def update(cls, app_id, value):
        cls.tasks_lock.acquire()
        if app_id in cls.tasks:
            cls.tasks[app_id] = value
        cls.tasks_lock.release()

    @classmethod
    def peek(cls, app_id):
        result = None
        cls.tasks_lock.acquire()
        if app_id in cls.tasks:
            result = cls.tasks[app_id]
        cls.tasks_lock.release()
        return result

    @classmethod
    def remove(cls, app_id):
        cls.tasks_lock.acquire()
        del cls.tasks[app_id]
        cls.tasks_lock.release()


class WorkerThread(StoppableThread):
    def __init__(self, pid, config):
        StoppableThread.__init__(self)
        Thread.__init__(self)
        self.pid = pid
        self.config = config

    def run(self):
        LOG = logging.getLogger("worker")
        LOG.info("Worker(%03d) start", self.pid)
        try:
            lpl = LitePipelineClient(self.config["manager_http_host"],
                                     self.config["manager_http_port"],
                                     user = self.config["manager_user"],
                                     password = self.config["manager_password"])
            while not self.stopped():
                try:
                    success = True
                    app_id = TasksCache.get()
                    if app_id:
                        LOG.info("downloading app_id: %s", app_id)
                        try:
                            tmp_path = os.path.join(self.config["data_path"], "tmp")
                            r = lpl.application_download(app_id, directory = tmp_path)
                            if r:
                                file_path, file_type = r
                                app_path = os.path.join(self.config["data_path"], "applications", app_id[:2], app_id[2:4], app_id)
                                if os.path.exists(app_path):
                                    shutil.rmtree(app_path)
                                os.makedirs(app_path)
                                shutil.copy2(file_path, os.path.join(app_path, "app.%s" % file_type))
                                os.remove(file_path)
                                if os.path.exists(os.path.join(app_path, "app")):
                                    shutil.rmtree(os.path.join(app_path, "app"))
                                if file_type == "tar.gz":
                                    t = tarfile.open(os.path.join(app_path, "app.tar.gz"), "r")
                                    t.extractall(app_path)
                                    path_parts = splitall(t.getnames()[0])
                                    app_root_name = path_parts[1] if path_parts[0] == "." else path_parts[0]
                                    t.close()
                                elif file_type == "zip":
                                    z = ZipFileWithPermissions(os.path.join(app_path, "app.zip"), "r")
                                    z.extractall(app_path)
                                    path_parts = splitall(z.namelist()[0])
                                    app_root_name = path_parts[1] if path_parts[0] == "." else path_parts[0]
                                    z.close()
                                app_config_path = os.path.join(app_path, app_root_name, "configuration.json")
                                f = open(app_config_path, "r")
                                app_config = json.loads(f.read())
                                f.close()
                                venvs = set()
                                for action in app_config["actions"]:
                                    if "env" in action and not is_uuid(action["env"]):
                                        venvs.add(action["env"])
                                for venv in list(venvs):
                                    venv_tar_path = os.path.join(app_path, app_root_name, "%s.tar.gz" % venv)
                                    if os.path.exists(venv_tar_path) and os.path.isfile(venv_tar_path): 
                                        venv_path = os.path.join(app_path, app_root_name, venv)
                                        if os.path.exists(venv_path):
                                            shutil.rmtree(venv_path)
                                        os.makedirs(venv_path)
                                        t = tarfile.open(venv_tar_path, "r")
                                        t.extractall(venv_path)
                                        t.close()
                                        update_venv_cfg(venv_path)
                                    else: # lose venv file
                                        success = False
                                        TasksCache.update(app_id, {"type": "error", "code": r.status_code, "message": "invalid application[%s] format" % app_id, "result": "invalid application[%s] format" % app_id})
                                        LOG.warning("invalid application[%s] format status: %s", app_id, r.status_code)
                                        break
                                os.rename(os.path.join(app_path, app_root_name), os.path.join(app_path, "app"))
                                if success:
                                    TasksCache.remove(app_id)
                        except OperationFailedError as e:
                            TasksCache.update(app_id, {"type": "error", "code": 400, "message": "download application[%s] failed" % app_id, "result": e})
                            LOG.warning("download[%s] message: %s", app_id, e)
                    else:
                        time.sleep(0.5)
                except Exception as e:
                    LOG.exception(e)
        except Exception as e:
            LOG.exception(e)
        LOG.info("Worker(%03d) exit", self.pid)


class Manager(Process):
    def __init__(self, pipe_client, worker_num, config):
        Process.__init__(self)
        self.pipe_client = pipe_client
        self.worker_num = worker_num
        self.config = config

    def run(self):
        logger.config_logging(file_name = "apps_manager.log",
                              log_level = "NOSET",
                              dir_name = self.config["log_path"],
                              when = "D",
                              interval = 1,
                              max_size = 20,
                              backup_count = 5,
                              console = True)
        LOG = logging.getLogger("manager")

        def sig_handler(sig, frame):
            LOG.warning("sig_handler Caught signal: %s", sig)

        LOG.info("Manager start")
        try:
            signal.signal(signal.SIGTERM, sig_handler)
            signal.signal(signal.SIGINT, sig_handler)

            threads = []
            for i in range(self.worker_num):
                t = WorkerThread(i, self.config)
                t.start()
                threads.append(t)

            lpl = LitePipelineClient(self.config["manager_http_host"],
                                     self.config["manager_http_port"],
                                     user = self.config["manager_user"],
                                     password = self.config["manager_password"])

            while True:
                LOG.debug("Manager main loop")
                command, app_id = self.pipe_client.recv()
                if command == Command.check_app:
                    ready = False
                    try:
                        LOG.debug("check app, app_id: %s", app_id)
                        r = lpl.application_info(app_id)
                        if r and "app_info" in r:
                            app_info = r["app_info"]
                            app_base_path = os.path.join(self.config["data_path"], "applications", app_id[:2], app_id[2:4], app_id)
                            app_tar_path = os.path.join(app_base_path, "app.tar.gz")
                            app_zip_path = os.path.join(app_base_path, "app.zip")
                            app_path = os.path.join(app_base_path, "app")
                            # check app.tar.gz
                            if os.path.exists(app_tar_path) and os.path.isfile(app_tar_path):
                                if app_info["sha1"] != file_sha1sum(app_tar_path):
                                    # download app.tar.gz && extract app.tar.gz
                                    pass
                                else:
                                    if not os.path.exists(app_path):
                                        # extract app.tar.gz
                                        pass
                                    else:
                                        ready = True
                            # check app.zip
                            if os.path.exists(app_zip_path) and os.path.isfile(app_zip_path):
                                if app_info["sha1"] != file_sha1sum(app_zip_path):
                                    # download app.zip && extract app.zip
                                    pass
                                else:
                                    if not os.path.exists(app_path):
                                        # extract app.tar.gz
                                        pass
                                    else:
                                        ready = True
                            # download app.tar.gz && extract app.tar.gz
                            if ready:
                                status = TasksCache.peek(app_id)
                                if status is not None:
                                    ready = False
                            else:
                                status = TasksCache.peek(app_id)
                                if isinstance(status, dict):
                                    ready = status
                                    TasksCache.remove(app_id)
                                elif status is None:
                                    TasksCache.set(app_id)
                    except Exception as e:
                        LOG.exception(e)
                        ready = {"type": "error", "message": "get app[%s] info failed: %s" % (app_id, str(e))}
                    self.pipe_client.send((command, ready))
                elif command == Command.exit:
                    for t in threads:
                        t.stop()
                    break
            for t in threads:
                t.join()
        except Exception as e:
            LOG.exception(e)
        LOG.info("Manager exit")


class ManagerClient(object):
    process_list = []
    process_dict = {}
    write_lock = locks.Lock()
    _instance = None

    def __init__(self, worker_num = 1):
        if ManagerClient._instance is None:
            self.worker_num = worker_num if worker_num > 0 else 1
            LOG.debug("ManagerClient, worker_num: %s", self.worker_num)
            pipe_master, pipe_client = Pipe()
            p = Manager(pipe_client, self.worker_num, CONFIG)
            p.daemon = True
            ManagerClient.process_list.append(p)
            ManagerClient.process_dict["manager"] = [p, pipe_master]
            p.start()
            ManagerClient._instance = self
        else:
            self.worker_num = ManagerClient._instance.worker_num

    @gen.coroutine
    def check_app(self, app_id):
        result = False
        LOG.debug("start check app, app_id: %s", app_id)
        with (yield ManagerClient.write_lock.acquire()):
            LOG.debug("get check app lock, app_id: %s", app_id)
            ManagerClient.process_dict["manager"][1].send((Command.check_app, app_id))
            LOG.debug("send check app, app_id: %s", app_id)
            while not ManagerClient.process_dict["manager"][1].poll():
                yield gen.moment
            LOG.debug("recv check app, app_id: %s", app_id)
            r = ManagerClient.process_dict["manager"][1].recv()
            LOG.debug("end check app, app_id: %s, r: %s", app_id, r)
        if r[1]:
            result = r[1]
        raise gen.Return(result)

    def close(self):
        try:
            LOG.info("close ManagerClient")
            ManagerClient.process_dict["manager"][1].send((Command.exit, None))
            for p in ManagerClient.process_list[1:]:
                p.terminate()
            for p in ManagerClient.process_list:
                while p.is_alive():
                    time.sleep(0.5)
                    LOG.debug("sleep 0.5 second")
            LOG.info("All Process Exit!")
        except Exception as e:
            LOG.exception(e)



# -*- coding: utf-8 -*-

import os
import sys
import time
import json
import hashlib
import logging
import datetime
import configparser
import threading
from threading import Thread, Lock
import zipfile
from zipfile import ZipFile, ZipInfo
from uuid import UUID

from tornado import ioloop
from tornado import gen

from litepipeline.node.config import CONFIG

LOG = logging.getLogger(__name__)

BUF_SIZE = 65536


class Servers(object):
    HTTP_SERVER = None
    DB_SERVERS = []
    TORNADO_INSTANCE = None


async def shutdown():
    LOG.info("Stopping Service(%s:%s)", CONFIG["http_host"], CONFIG["http_port"])
    if Servers.HTTP_SERVER:
        Servers.HTTP_SERVER.stop()
        LOG.info("Stop http server!")
    for db_server in Servers.DB_SERVERS:
        db_server.close()
        LOG.info("Stop db server!")
    await gen.sleep(1)
    LOG.info("Will shutdown ...")
    ioloop.IOLoop.current().stop()


def sig_handler(sig, frame):
    LOG.warning("sig_handler Caught signal: %s", sig)
    ioloop.IOLoop.current().add_callback_from_signal(shutdown)


class Errors(object):
    OK = "ok"
    errors = {
        "ServerException": {"name": "ServerException", "message": "server exception"},
        "InvalidParameters": {"name": "InvalidParameters", "message": "invalid parameters"},
        "OperationFailed": {"name": "OperationFailed", "message": "operation failed"},
        "AppNotExists": {"name": "AppNotExists", "message": "application not exists"},
        "TaskNotExists": {"name": "TaskNotExists", "message": "task not exists"},
        "OperationRunning": {"name": "OperationRunning", "message": "operation running"},
        "WorkspaceNotExists": {"name": "WorkspaceNotExists", "message": "workspace not exists"},
        "WorkspaceNotPacked": {"name": "WorkspaceNotPacked", "message": "workspace not packed"},
    }

    @classmethod
    def set_result_error(cls, error_name, result, message = ""):
        if error_name in cls.errors:
            result["result"] = error_name
            if message == "":
                result["message"] = cls.errors[error_name]["message"]
            else:
                result["message"] = message
        else:
            result["result"] = "UnknownError"
            result["message"] = "unknown error"


class Stage(object):
    pending = "pending"
    running = "running"
    finished = "finished"


class Status(object):
    fail = "fail"
    success = "success"
    kill = "kill"
    cancel = "cancel"
    terminate = "terminate"


class Signal(object):
    kill = -9
    terminate = -15
    stop = -15
    cancel = -50


class JSONLoadError(Exception):
    def __init__(self, message):
        self.message = message


class MetaNotDictError(Exception):
    def __init__(self, message):
        self.message = message


def file_sha1sum(file_path):
    sha1 = hashlib.sha1()
    with open(file_path, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()


def file_md5sum(file_path):
    md5 = hashlib.md5()
    with open(file_path, 'rb') as fp:
        while True:
            data = fp.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


def splitall(path):
    allparts = []
    while True:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts


def has_parent_directory(parent, paths):
    result = True
    for p in paths:
        if not p.startswith(parent):
            result = False
            break
    return result


def get_workspace_path(create_at, task_id = None, action_name = None, service_id = None, config = None):
    result = ""
    date_create_at = datetime.datetime.strptime(create_at, "%Y-%m-%d %H:%M:%S.%f")
    date_directory_name = date_create_at.strftime("%Y-%m-%d")
    data_path = CONFIG["data_path"] if config is None else config["data_path"]
    if service_id:
        result = os.path.join(data_path, "tmp", "service_workspace", service_id[:2], service_id[2:4], service_id, action_name)
    elif task_id and action_name:
        result = os.path.join(data_path, "tmp", "workspace", date_directory_name, task_id[:2], task_id[2:4], task_id, action_name)
    elif task_id:
        result = os.path.join(data_path, "tmp", "workspace", date_directory_name, task_id[:2], task_id[2:4], task_id)
    else:
        result = os.path.join(data_path, "tmp", "workspace", date_directory_name)
    return result


def get_download_path(task_id, action_name, service_id, config = None):
    result = ""
    data_path = CONFIG["data_path"] if config is None else config["data_path"]
    if service_id and action_name:
        result = os.path.join(data_path, "tmp", "download", "%s.%s.tar.gz" % (service_id, action_name))
    elif task_id and action_name:
        result = os.path.join(data_path, "tmp", "download", "%s.%s.tar.gz" % (task_id, action_name))
    else:
        result = os.path.join(data_path, "tmp", "download")
    return result


class FakeSectionHead(object):
    def __init__(self, fp):
        self.fp = fp
        self.sechead = '[global]\n'

    def readline(self):
        if self.sechead:
            try: 
                return self.sechead
            finally: 
                self.sechead = None
        else: 
            return self.fp.readline()

    def __iter__(self):
        line = self.readline()
        while line:
            yield line
            line = self.readline()

    def close(self):
        self.fp.close()


def update_venv_cfg(venv_path):
    venv_path = venv_path
    cfg_path = os.path.join(venv_path, "pyvenv.cfg")
    config = configparser.ConfigParser()
    items = []
    if os.path.exists(cfg_path) and os.path.isfile(cfg_path):
        fp = FakeSectionHead(open(cfg_path))
        config.read_file(fp)
        fp.close()
        items = config.items("global")
        LOG.debug(items)
        with open(cfg_path, "w") as fp:
            for item in items:
                if item[0] == "home":
                    python_home_path = item[1]
                    if sys.base_prefix:
                        python_home_path = os.path.join(sys.base_prefix, "bin")
                    elif sys.base_exec_prefix:
                        python_home_path = os.path.join(sys.base_exec_prefix, "bin")
                    fp.write("%s = %s\n" % ("home", python_home_path))
                else:
                    fp.write("%s = %s\n" % item)


class StoppableThread(Thread):
    """
    Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition.
    """

    def __init__(self):
        super(StoppableThread, self).__init__()
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.isSet()


class ZipFileWithPermissions(ZipFile):
    '''
    Custom ZipFile class handling file permissions
    '''

    def _extract_member(self, member, targetpath, pwd):
        if not isinstance(member, ZipInfo):
            member = self.getinfo(member)

        targetpath = super()._extract_member(member, targetpath, pwd)

        attr = member.external_attr >> 16
        if attr != 0:
            os.chmod(targetpath, attr)
        return targetpath


def is_uuid(u, version = 4):
    result = True
    try:
        UUID(u, version = version)
    except Exception as e:
        result = False
    return result


def init_storage():
    directories = [
        os.path.join(CONFIG["data_path"], "venvs"),
        os.path.join(CONFIG["data_path"], "applications"),
        os.path.join(CONFIG["data_path"], "tmp", "service_workspace"),
        os.path.join(CONFIG["data_path"], "tmp", "workspace"),
        os.path.join(CONFIG["data_path"], "tmp", "download"),
    ]
    for d in directories:
        if not os.path.exists(d) or not os.path.isdir(d):
            os.makedirs(d)



# -*- coding: utf-8 -*-

import os
import sys
import time
import json
import logging
import shutil
import tarfile
import zipfile
from zipfile import ZipFile, ZipInfo
import signal
from pathlib import Path
import threading
from threading import Thread, Lock
from multiprocessing import Process, Queue, Pipe

from tornado import gen, locks
import requests
from litepipeline_helper.models.client import LitePipelineClient

from litepipeline.node.utils.common import file_sha1sum, splitall, update_venv_cfg, StoppableThread, ZipFileWithPermissions, has_parent_directory
from litepipeline.node.config import CONFIG
from litepipeline import logger

LOG = logging.getLogger(__name__)


class Command(object):
    check_venv = "check_venv"
    exit = "exit"


class TasksCache(object):
    tasks = {}
    tasks_lock = Lock()

    @classmethod
    def set(cls, venv_id):
        cls.tasks_lock.acquire()
        if venv_id not in cls.tasks:
            cls.tasks[venv_id] = False
        cls.tasks_lock.release()

    @classmethod
    def get(cls):
        result = False
        cls.tasks_lock.acquire()
        try:
            for venv_id in cls.tasks:
                if not cls.tasks[venv_id]:
                    cls.tasks[venv_id] = True
                    result = venv_id
                    break
        except Exception as e:
            LOG.exception(e)
        cls.tasks_lock.release()
        return result

    @classmethod
    def update(cls, venv_id, value):
        cls.tasks_lock.acquire()
        if venv_id in cls.tasks:
            cls.tasks[venv_id] = value
        cls.tasks_lock.release()

    @classmethod
    def peek(cls, venv_id):
        result = None
        cls.tasks_lock.acquire()
        if venv_id in cls.tasks:
            result = cls.tasks[venv_id]
        cls.tasks_lock.release()
        return result

    @classmethod
    def remove(cls, venv_id):
        cls.tasks_lock.acquire()
        del cls.tasks[venv_id]
        cls.tasks_lock.release()


class WorkerThread(StoppableThread):
    def __init__(self, pid, config):
        StoppableThread.__init__(self)
        Thread.__init__(self)
        self.pid = pid
        self.config = config

    def run(self):
        LOG = logging.getLogger("worker")
        LOG.info("Worker(%03d) start", self.pid)
        try:
            lpl = LitePipelineClient(self.config["manager_http_host"],
                                     self.config["manager_http_port"],
                                     user = self.config["manager_user"],
                                     password = self.config["manager_password"])
            while not self.stopped():
                try:
                    success = True
                    venv_id = TasksCache.get()
                    if venv_id:
                        LOG.info("downloading venv_id: %s", venv_id)
                        try:
                            tmp_path = os.path.join(self.config["data_path"], "tmp")
                            r = lpl.venv_download(venv_id, directory = tmp_path)
                            if r:
                                file_path, file_type = r
                                venv_path = os.path.join(self.config["data_path"], "venvs", venv_id[:2], venv_id[2:4], venv_id)
                                if os.path.exists(venv_path):
                                    shutil.rmtree(venv_path)
                                os.makedirs(venv_path)
                                shutil.copy2(file_path, os.path.join(venv_path, "venv.%s" % file_type))
                                os.remove(file_path)
                                if os.path.exists(os.path.join(venv_path, "venv")):
                                    shutil.rmtree(os.path.join(venv_path, "venv"))
                                if file_type == "tar.gz":
                                    t = tarfile.open(os.path.join(venv_path, "venv.tar.gz"), "r")
                                    sub_paths = t.getnames()
                                    path_parts = splitall(sub_paths[0])
                                    venv_root_name = path_parts[1] if path_parts[0] == "." else path_parts[0]
                                    if has_parent_directory(venv_root_name, sub_paths):
                                        t.extractall(venv_path)
                                        t.close()
                                    else:
                                        t.extractall(os.path.join(venv_path, "venv"))
                                        venv_root_name = "venv"
                                        t.close()
                                elif file_type == "zip":
                                    z = ZipFileWithPermissions(os.path.join(venv_path, "venv.zip"), "r")
                                    sub_paths = z.namelist()
                                    path_parts = splitall(sub_paths[0])
                                    venv_root_name = path_parts[1] if path_parts[0] == "." else path_parts[0]
                                    if has_parent_directory(venv_root_name, sub_paths):
                                        z.extractall(venv_path)
                                        z.close()
                                    else:
                                        z.extractall(os.path.join(venv_path, "venv"))
                                        venv_root_name = "venv"
                                        z.close()
                                if venv_root_name:
                                    os.rename(os.path.join(venv_path, venv_root_name), os.path.join(venv_path, "venv"))
                                    update_venv_cfg(os.path.join(venv_path, "venv"))
                                    TasksCache.remove(venv_id)
                                else:
                                    TasksCache.update(venv_id, {"type": "error", "code": r.status_code, "message": "invalid venv[%s] format" % venv_id, "result": "invalid venv[%s] format" % venv_id})
                                    LOG.warning("invalid venv[%s] format status: %s", venv_id, r.status_code)
                            elif r.status_code == 400:
                                TasksCache.update(venv_id, {"type": "error", "code": r.status_code, "message": "download venv[%s] failed" % venv_id, "result": r.json()})
                                LOG.warning("download[%s] status: %s", venv_id, r.status_code)
                            else:
                                TasksCache.update(venv_id, {"type": "error", "code": r.status_code, "message": "download venv[%s] failed" % venv_id})
                                LOG.warning("download[%s] status: %s", venv_id, r.status_code)
                        except OperationFailedError as e:
                            TasksCache.update(app_id, {"type": "error", "code": 400, "message": "download venv[%s] failed" % venv_id, "result": e})
                            LOG.warning("download[%s] message: %s", venv_id, e)
                    else:
                        time.sleep(0.5)
                except Exception as e:
                    LOG.exception(e)
        except Exception as e:
            LOG.exception(e)
        LOG.info("Worker(%03d) exit", self.pid)


class Manager(Process):
    def __init__(self, pipe_client, worker_num, config):
        Process.__init__(self)
        self.pipe_client = pipe_client
        self.worker_num = worker_num
        self.config = config

    def run(self):
        logger.config_logging(file_name = "venvs_manager.log",
                              log_level = "NOSET",
                              dir_name = self.config["log_path"],
                              when = "D",
                              interval = 1,
                              max_size = 20,
                              backup_count = 5,
                              console = True)
        LOG = logging.getLogger("manager")

        def sig_handler(sig, frame):
            LOG.warning("sig_handler Caught signal: %s", sig)

        LOG.info("Manager start")
        try:
            signal.signal(signal.SIGTERM, sig_handler)
            signal.signal(signal.SIGINT, sig_handler)

            threads = []
            for i in range(self.worker_num):
                t = WorkerThread(i, self.config)
                t.start()
                threads.append(t)

            lpl = LitePipelineClient(self.config["manager_http_host"],
                                     self.config["manager_http_port"],
                                     user = self.config["manager_user"],
                                     password = self.config["manager_password"])

            while True:
                LOG.debug("Manager main loop")
                command, venv_id = self.pipe_client.recv()
                if command == Command.check_venv:
                    ready = False
                    try:
                        LOG.debug("check venv, venv_id: %s", venv_id)
                        r = lpl.venv_info(venv_id)
                        if r and "venv_info" in r:
                            venv_info = r["venv_info"]
                            venv_base_path = os.path.join(self.config["data_path"], "venvs", venv_id[:2], venv_id[2:4], venv_id)
                            venv_tar_path = os.path.join(venv_base_path, "venv.tar.gz")
                            venv_zip_path = os.path.join(venv_base_path, "venv.zip")
                            venv_path = os.path.join(venv_base_path, "venv")
                            # check venv.tar.gz
                            if os.path.exists(venv_tar_path) and os.path.isfile(venv_tar_path):
                                if venv_info["sha1"] != file_sha1sum(venv_tar_path):
                                    # download venv.tar.gz && extract venv.tar.gz
                                    pass
                                else:
                                    if not os.path.exists(venv_path):
                                        # extract venv.tar.gz
                                        pass
                                    else:
                                        ready = True
                            # check venv.zip
                            if os.path.exists(venv_zip_path) and os.path.isfile(venv_zip_path):
                                if venv_info["sha1"] != file_sha1sum(venv_zip_path):
                                    # download venv.zip && extract venv.zip
                                    pass
                                else:
                                    if not os.path.exists(venv_path):
                                        # extract venv.tar.gz
                                        pass
                                    else:
                                        ready = True
                            # download venv.tar.gz && extract venv.tar.gz
                            if ready:
                                status = TasksCache.peek(venv_id)
                                if status is not None:
                                    ready = False
                            else:
                                status = TasksCache.peek(venv_id)
                                if isinstance(status, dict):
                                    ready = status
                                    TasksCache.remove(venv_id)
                                elif status is None:
                                    TasksCache.set(venv_id)
                    except Exception as e:
                        LOG.exception(e)
                        ready = {"type": "error", "message": "get venv[%s] info failed: %s" % (venv_id, str(e))}
                    self.pipe_client.send((command, ready))
                elif command == Command.exit:
                    for t in threads:
                        t.stop()
                    break
            for t in threads:
                t.join()
        except Exception as e:
            LOG.exception(e)
        LOG.info("Manager exit")


class ManagerClient(object):
    process_list = []
    process_dict = {}
    write_lock = locks.Lock()
    _instance = None

    def __init__(self, worker_num = 1):
        if ManagerClient._instance is None:
            self.worker_num = worker_num if worker_num > 0 else 1
            LOG.debug("ManagerClient, worker_num: %s", self.worker_num)
            pipe_master, pipe_client = Pipe()
            p = Manager(pipe_client, self.worker_num, CONFIG)
            p.daemon = True
            ManagerClient.process_list.append(p)
            ManagerClient.process_dict["manager"] = [p, pipe_master]
            p.start()
            ManagerClient._instance = self
        else:
            self.worker_num = ManagerClient._instance.worker_num

    @gen.coroutine
    def check_venv(self, venv_id):
        result = False
        LOG.debug("start check venv, venv_id: %s", venv_id)
        with (yield ManagerClient.write_lock.acquire()):
            LOG.debug("get check venv lock, venv_id: %s", venv_id)
            ManagerClient.process_dict["manager"][1].send((Command.check_venv, venv_id))
            LOG.debug("send check venv, venv_id: %s", venv_id)
            while not ManagerClient.process_dict["manager"][1].poll():
                yield gen.moment
            LOG.debug("recv check venv, venv_id: %s", venv_id)
            r = ManagerClient.process_dict["manager"][1].recv()
            LOG.debug("end check venv, venv_id: %s, r: %s", venv_id, r)
        if r[1]:
            result = r[1]
        raise gen.Return(result)

    def close(self):
        try:
            LOG.info("close ManagerClient")
            ManagerClient.process_dict["manager"][1].send((Command.exit, None))
            for p in ManagerClient.process_list[1:]:
                p.terminate()
            for p in ManagerClient.process_list:
                while p.is_alive():
                    time.sleep(0.5)
                    LOG.debug("sleep 0.5 second")
            LOG.info("All Process Exit!")
        except Exception as e:
            LOG.exception(e)



# -*- coding: utf-8 -*-

import os
import io
import time
import json
import logging
import shutil
import tarfile
import zipfile

from litedfs_client.client import LiteDFSClient

from litepipeline.manager.models.venvs import Venvs
from litepipeline.manager.models.venv_history import VenvHistory
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, splitall
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class VenvManagerBase(object):
    def __init__(self):
        pass

    def create(self, name, description, source_path):
        pass

    def update(self, venv_id, name, description, source_path):
        pass

    def list(self, offset, limit, filters = {}):
        pass

    def list_history(self, offset, limit, filters = {}):
        pass

    def info(self, venv_id):
        pass

    def info_history(self, history_id, venv_id = ""):
        pass

    def activate_history(self, history_id, venv_id = ""):
        pass

    def delete(self, venv_id):
        pass

    def delete_history(self, history_id, venv_id):
        pass

    def get_venv_config(self, venv_id, sha1):
        pass

    def open(self, venv_id, sha1):
        pass

    def close(self):
        pass


class VenvLocalTarGzManager(VenvManagerBase):
    _instance = None
    name = "VenvLocalTarGzManager"

    def __new__(cls):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.root_path = CONFIG["data_path"]
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def make_venv_path(self, venv_id):
        return os.path.join(self.root_path, "venvs", venv_id[:2], venv_id[2:4], venv_id)

    def make_venv_version_path(self, venv_id, sha1):
        return os.path.join(self.make_venv_path(venv_id), sha1)

    def create(self, name, description, source_path):
        sha1 = file_sha1sum(source_path)
        LOG.debug("sha1: %s, %s", sha1, type(sha1))
        venv_id = Venvs.instance().add(name, sha1, description = description)
        venv_path = self.make_venv_version_path(venv_id, sha1)
        if os.path.exists(venv_path):
            shutil.rmtree(venv_path)
        os.makedirs(venv_path)
        shutil.copy2(source_path, os.path.join(venv_path, "venv.tar.gz"))
        os.remove(source_path)
        VenvHistory.instance().add(venv_id, sha1, description = description)
        return venv_id

    def update(self, venv_id, name, description, source_path):
        result = True
        try:
            data = {}
            need_update = False
            if name:
                data["name"] = name
            if description:
                data["description"] = description
            if os.path.exists(source_path) and os.path.isfile(source_path):
                sha1 = file_sha1sum(source_path)
                data["sha1"] = sha1
                LOG.debug("sha1: %s, %s", sha1, type(sha1))
                venv_path = self.make_venv_version_path(venv_id, sha1)
                if os.path.exists(venv_path):
                    shutil.rmtree(venv_path)
                os.makedirs(venv_path)
                shutil.copy2(source_path, os.path.join(venv_path, "venv.tar.gz"))
                os.remove(source_path)
                need_update = True
            if data or need_update:
                success = Venvs.instance().update(venv_id, data)
                if success:
                    if "sha1" in data:
                        description = "" if "description" not in data else data["description"]
                        VenvHistory.instance().add(venv_id, data["sha1"], description = description)
                else:
                    result = False
        except Exception as e:
            LOG.exception(e)
        return result

    def list(self, offset, limit, filters = {}):
        return Venvs.instance().list(offset = offset, limit = limit, filters = filters)

    def list_history(self, offset, limit, filters = {}):
        return VenvHistory.instance().list(offset = offset, limit = limit, filters = filters)

    def info(self, venv_id):
        return Venvs.instance().get(venv_id)

    def info_history(self, history_id, venv_id = ""):
        return VenvHistory.instance().get(history_id, venv_id = venv_id)

    def activate_history(self, history_id, venv_id = ""):
        result = False
        try:
            if venv_id:
                history = self.info_history(history_id, venv_id = venv_id)
                if history:
                    data = {"sha1": history["sha1"]}
                    if history["description"]:
                        data["description"] = history["description"]
                    success = Venvs.instance().update(venv_id, data)
                    if success:
                        result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete(self, venv_id):
        result = False
        try:
            success = Venvs.instance().delete(venv_id)
            if success:
                VenvHistory.instance().delete_by_venv_id(venv_id)
                venv_path = self.make_venv_path(venv_id)
                if os.path.exists(venv_path):
                    shutil.rmtree(venv_path)
                    LOG.debug("remove directory: %s", venv_path)
                result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete_history(self, history_id, venv_id):
        result = False
        try:
            history = VenvHistory.instance().delete_by_history_id_venv_id(history_id, venv_id)
            if history and history is not None:
                filters = VenvHistory.instance().parse_filters({"venv_id": history["venv_id"], "sha1": history["sha1"]})
                num = VenvHistory.instance().count(filters)
                if num == 0:
                    venv_path = self.make_venv_version_path(history["venv_id"], history["sha1"])
                    if os.path.exists(venv_path):
                        shutil.rmtree(venv_path)
                        LOG.debug("remove directory: %s", venv_path)
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def open(self, venv_id, sha1):
        result = False
        try:
            venv_path = os.path.join(self.make_venv_version_path(venv_id, sha1), "venv.tar.gz")
            if os.path.exists(venv_path) and os.path.isfile(venv_path):
                result = open(venv_path, "rb")
        except Exception as e:
            LOG.exception(e)
        return result

    def close(self):
        pass


class VenvLocalZipManager(VenvManagerBase):
    _instance = None
    name = "VenvLocalZipManager"

    def __new__(cls):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.root_path = CONFIG["data_path"]
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def make_venv_path(self, venv_id):
        return os.path.join(self.root_path, "venvs", venv_id[:2], venv_id[2:4], venv_id)

    def make_venv_version_path(self, venv_id, sha1):
        return os.path.join(self.make_venv_path(venv_id), sha1)

    def create(self, name, description, source_path):
        sha1 = file_sha1sum(source_path)
        LOG.debug("sha1: %s, %s", sha1, type(sha1))
        venv_id = Venvs.instance().add(name, sha1, description = description)
        venv_path = self.make_venv_version_path(venv_id, sha1)
        if os.path.exists(venv_path):
            shutil.rmtree(venv_path)
        os.makedirs(venv_path)
        shutil.copy2(source_path, os.path.join(venv_path, "venv.zip"))
        os.remove(source_path)
        VenvHistory.instance().add(venv_id, sha1, description = description)
        return venv_id

    def update(self, venv_id, name, description, source_path):
        result = True
        try:
            data = {}
            need_update = False
            if name:
                data["name"] = name
            if description:
                data["description"] = description
            if os.path.exists(source_path) and os.path.isfile(source_path):
                sha1 = file_sha1sum(source_path)
                data["sha1"] = sha1
                LOG.debug("sha1: %s, %s", sha1, type(sha1))
                venv_path = self.make_venv_version_path(venv_id, sha1)
                if os.path.exists(venv_path):
                    shutil.rmtree(venv_path)
                os.makedirs(venv_path)
                shutil.copy2(source_path, os.path.join(venv_path, "venv.zip"))
                os.remove(source_path)
                need_update = True
            if data or need_update:
                success = Venvs.instance().update(venv_id, data)
                if success:
                    if "sha1" in data:
                        description = "" if "description" not in data else data["description"]
                        VenvHistory.instance().add(venv_id, data["sha1"], description = description)
                else:
                    result = False
        except Exception as e:
            LOG.exception(e)
        return result

    def list(self, offset, limit, filters = {}):
        return Venvs.instance().list(offset = offset, limit = limit, filters = filters)

    def list_history(self, offset, limit, filters = {}):
        return VenvHistory.instance().list(offset = offset, limit = limit, filters = filters)

    def info(self, venv_id):
        return Venvs.instance().get(venv_id)

    def info_history(self, history_id, venv_id = ""):
        return VenvHistory.instance().get(history_id, venv_id = venv_id)

    def activate_history(self, history_id, venv_id = ""):
        result = False
        try:
            if venv_id:
                history = self.info_history(history_id, venv_id = venv_id)
                if history:
                    data = {"sha1": history["sha1"]}
                    if history["description"]:
                        data["description"] = history["description"]
                    success = Venvs.instance().update(venv_id, data)
                    if success:
                        result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete(self, venv_id):
        result = False
        try:
            success = Venvs.instance().delete(venv_id)
            if success:
                VenvHistory.instance().delete_by_venv_id(venv_id)
                venv_path = self.make_venv_path(venv_id)
                if os.path.exists(venv_path):
                    shutil.rmtree(venv_path)
                    LOG.debug("remove directory: %s", venv_path)
                result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete_history(self, history_id, venv_id):
        result = False
        try:
            history = VenvHistory.instance().delete_by_history_id_venv_id(history_id, venv_id)
            if history and history is not None:
                filters = VenvHistory.instance().parse_filters({"venv_id": history["venv_id"], "sha1": history["sha1"]})
                num = VenvHistory.instance().count(filters)
                if num == 0:
                    venv_path = self.make_venv_version_path(history["venv_id"], history["sha1"])
                    if os.path.exists(venv_path):
                        shutil.rmtree(venv_path)
                        LOG.debug("remove directory: %s", venv_path)
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def open(self, venv_id, sha1):
        result = False
        try:
            venv_path = os.path.join(self.make_venv_version_path(venv_id, sha1), "venv.zip")
            if os.path.exists(venv_path) and os.path.isfile(venv_path):
                result = open(venv_path, "rb")
        except Exception as e:
            LOG.exception(e)
        return result

    def close(self):
        pass


class VenvLDFSTarGzManager(VenvManagerBase):
    _instance = None
    name = "VenvLDFSTarGzManager"

    def __new__(cls):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.root_path = os.path.join("/litemanager_%s:%s" % (CONFIG["http_host"], CONFIG["http_port"]), "data")
            cls._instance.ldfs = LiteDFSClient(CONFIG["ldfs_http_host"], CONFIG["ldfs_http_port"])
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def make_venv_path(self, venv_id):
        return os.path.join(self.root_path, "venvs", venv_id[:2], venv_id[2:4], venv_id)

    def make_venv_version_path(self, venv_id, sha1):
        return os.path.join(self.make_venv_path(venv_id), sha1)

    def create(self, name, description, source_path):
        sha1 = file_sha1sum(source_path)
        LOG.debug("sha1: %s, %s", sha1, type(sha1))
        venv_id = Venvs.instance().add(name, sha1, description = description)
        venv_path = self.make_venv_version_path(venv_id, sha1)
        self.ldfs.delete_directory(venv_path)
        self.ldfs.create_file(source_path, os.path.join(venv_path, "venv.tar.gz"), replica = 1)
        os.remove(source_path)
        VenvHistory.instance().add(venv_id, sha1, description = description)
        return venv_id

    def update(self, venv_id, name, description, source_path):
        result = True
        try:
            data = {}
            need_update = False
            if name:
                data["name"] = name
            if description:
                data["description"] = description
            if os.path.exists(source_path) and os.path.isfile(source_path):
                sha1 = file_sha1sum(source_path)
                data["sha1"] = sha1
                LOG.debug("sha1: %s, %s", sha1, type(sha1))
                venv_path = self.make_venv_version_path(venv_id, sha1)
                self.ldfs.delete_directory(venv_path)
                self.ldfs.create_file(source_path, os.path.join(venv_path, "venv.tar.gz"), replica = 1)
                os.remove(source_path)
                need_update = True
            if data or need_update:
                success = Venvs.instance().update(venv_id, data)
                if success:
                    if "sha1" in data:
                        description = "" if "description" not in data else data["description"]
                        VenvHistory.instance().add(venv_id, data["sha1"], description = description)
                else:
                    result = False
        except Exception as e:
            LOG.exception(e)
        return result

    def list(self, offset, limit, filters = {}):
        return Venvs.instance().list(offset = offset, limit = limit, filters = filters)

    def list_history(self, offset, limit, filters = {}):
        return VenvHistory.instance().list(offset = offset, limit = limit, filters = filters)

    def info(self, venv_id):
        return Venvs.instance().get(venv_id)

    def info_history(self, history_id, venv_id = ""):
        return VenvHistory.instance().get(history_id, venv_id = venv_id)

    def activate_history(self, history_id, venv_id = ""):
        result = False
        try:
            if venv_id:
                history = self.info_history(history_id, venv_id = venv_id)
                if history:
                    data = {"sha1": history["sha1"]}
                    if history["description"]:
                        data["description"] = history["description"]
                    success = Venvs.instance().update(venv_id, data)
                    if success:
                        result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete(self, venv_id):
        result = False
        try:
            success = Venvs.instance().delete(venv_id)
            if success:
                VenvHistory.instance().delete_by_venv_id(venv_id)
                venv_path = self.make_venv_path(venv_id)
                self.ldfs.delete_directory(venv_path)
                LOG.debug("remove ldfs directory: %s", venv_path)
                result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete_history(self, history_id, venv_id):
        result = False
        try:
            history = VenvHistory.instance().delete_by_history_id_venv_id(history_id, venv_id)
            if history and history is not None:
                filters = VenvHistory.instance().parse_filters({"venv_id": history["venv_id"], "sha1": history["sha1"]})
                num = VenvHistory.instance().count(filters)
                if num == 0:
                    venv_path = self.make_venv_version_path(history["venv_id"], history["sha1"])
                    self.ldfs.delete_directory(venv_path)
                    LOG.debug("remove ldfs directory: %s", venv_path)
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def open(self, venv_id, sha1):
        result = False
        try:
            venv_path = os.path.join(self.make_venv_version_path(venv_id, sha1), "venv.tar.gz")
            remote_file = self.ldfs.open_remote_file(venv_path)
            if remote_file:
                result = io.BufferedReader(remote_file, buffer_size = 1024 * 1024)
        except Exception as e:
            LOG.exception(e)
        return result

    def close(self):
        pass


class VenvLDFSZipManager(VenvManagerBase):
    _instance = None
    name = "VenvLDFSZipManager"

    def __new__(cls):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.root_path = os.path.join("/litemanager_%s:%s" % (CONFIG["http_host"], CONFIG["http_port"]), "data")
            cls._instance.ldfs = LiteDFSClient(CONFIG["ldfs_http_host"], CONFIG["ldfs_http_port"])
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def make_venv_path(self, venv_id):
        return os.path.join(self.root_path, "venvs", venv_id[:2], venv_id[2:4], venv_id)

    def make_venv_version_path(self, venv_id, sha1):
        return os.path.join(self.make_venv_path(venv_id), sha1)

    def create(self, name, description, source_path):
        sha1 = file_sha1sum(source_path)
        LOG.debug("sha1: %s, %s", sha1, type(sha1))
        venv_id = Venvs.instance().add(name, sha1, description = description)
        venv_path = self.make_venv_version_path(venv_id, sha1)
        self.ldfs.delete_directory(venv_path)
        self.ldfs.create_file(source_path, os.path.join(venv_path, "venv.zip"), replica = 1)
        os.remove(source_path)
        VenvHistory.instance().add(venv_id, sha1, description = description)
        return venv_id

    def update(self, venv_id, name, description, source_path):
        result = True
        try:
            data = {}
            need_update = False
            if name:
                data["name"] = name
            if description:
                data["description"] = description
            if os.path.exists(source_path) and os.path.isfile(source_path):
                sha1 = file_sha1sum(source_path)
                data["sha1"] = sha1
                LOG.debug("sha1: %s, %s", sha1, type(sha1))
                venv_path = self.make_venv_version_path(venv_id, sha1)
                self.ldfs.delete_directory(venv_path)
                self.ldfs.create_file(source_path, os.path.join(venv_path, "venv.zip"), replica = 1)
                os.remove(source_path)
                need_update = True
            if data or need_update:
                success = Venvs.instance().update(venv_id, data)
                if success:
                    if "sha1" in data:
                        description = "" if "description" not in data else data["description"]
                        VenvHistory.instance().add(venv_id, data["sha1"], description = description)
                else:
                    result = False
        except Exception as e:
            LOG.exception(e)
        return result

    def list(self, offset, limit, filters = {}):
        return Venvs.instance().list(offset = offset, limit = limit, filters = filters)

    def list_history(self, offset, limit, filters = {}):
        return VenvHistory.instance().list(offset = offset, limit = limit, filters = filters)

    def info(self, venv_id):
        return Venvs.instance().get(venv_id)

    def info_history(self, history_id, venv_id = ""):
        return VenvHistory.instance().get(history_id, venv_id = venv_id)

    def activate_history(self, history_id, venv_id = ""):
        result = False
        try:
            if venv_id:
                history = self.info_history(history_id, venv_id = venv_id)
                if history:
                    data = {"sha1": history["sha1"]}
                    if history["description"]:
                        data["description"] = history["description"]
                    success = Venvs.instance().update(venv_id, data)
                    if success:
                        result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete(self, venv_id):
        result = False
        try:
            success = Venvs.instance().delete(venv_id)
            if success:
                VenvHistory.instance().delete_by_venv_id(venv_id)
                venv_path = self.make_venv_path(venv_id)
                self.ldfs.delete_directory(venv_path)
                LOG.debug("remove ldfs directory: %s", venv_path)
                result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete_history(self, history_id, venv_id):
        result = False
        try:
            history = VenvHistory.instance().delete_by_history_id_venv_id(history_id, venv_id)
            if history and history is not None:
                filters = VenvHistory.instance().parse_filters({"venv_id": history["venv_id"], "sha1": history["sha1"]})
                num = VenvHistory.instance().count(filters)
                if num == 0:
                    venv_path = self.make_venv_version_path(history["venv_id"], history["sha1"])
                    self.ldfs.delete_directory(venv_path)
                    LOG.debug("remove ldfs directory: %s", venv_path)
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def open(self, venv_id, sha1):
        result = False
        try:
            venv_path = os.path.join(self.make_venv_version_path(venv_id, sha1), "venv.zip")
            remote_file = self.ldfs.open_remote_file(venv_path)
            if remote_file:
                result = io.BufferedReader(remote_file, buffer_size = 1024 * 1024)
        except Exception as e:
            LOG.exception(e)
        return result

    def close(self):
        pass


class VenvManager(object):
    _instance = None
    name = "VenvManager"

    def __new__(cls):
        if not cls._instance:
            if "venv_store" in CONFIG:
                if CONFIG["venv_store"] == "local.tar.gz":
                    cls._instance = VenvLocalTarGzManager()
                elif CONFIG["venv_store"] == "local.zip":
                    cls._instance = VenvLocalZipManager()
                elif CONFIG["venv_store"] == "ldfs.tar.gz":
                    cls._instance = VenvLDFSTarGzManager()
                elif CONFIG["venv_store"] == "ldfs.zip":
                    cls._instance = VenvLDFSZipManager()
                else:
                    cls._instance = VenvLocalTarGzManager()
            else:
                cls._instance = VenvLocalTarGzManager()
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance



# -*- coding: utf-8 -*-

import os
import time
import json
import logging
import shutil
import tarfile

import requests
from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, StreamBaseHandler, auth_check
from litepipeline.manager.utils.venv_manager import VenvManager
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, Errors, splitall
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class CreateVenvHandler(StreamBaseHandler):
    @auth_check
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            name = self.get_form_argument("name", "")
            description = self.get_form_argument("description", "")
            file_path = self.file_path.decode("utf-8")
            if name and os.path.exists(file_path) and os.path.isfile(file_path):
                file_name = os.path.split(file_path)[-1].lower()
                if ((CONFIG["venv_store"].endswith("tar.gz") and file_name.endswith("tar.gz")) or
                    (CONFIG["venv_store"].endswith("zip") and file_name.endswith("zip"))):
                    result["venv_id"] = VenvManager.instance().create(name, description, file_path)
                else:
                    LOG.warning("venv wrong format")
                    Errors.set_result_error("VenvWrongFormat", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        LOG.debug("deploy venv package use: %ss", time.time() - self.start)
        self.write(result)
        self.finish()


class ListVenvHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            filters = {}
            name = self.get_argument("name", "")
            if name:
                filters["name"] = name
            venv_id = self.get_argument("id", "")
            if venv_id:
                filters["id"] = venv_id
            LOG.debug("ListVenvHandler offset: %s, limit: %s", offset, limit)
            r = VenvManager.instance().list(offset, limit, filters = filters)
            result["venvs"] = r["venvs"]
            result["total"] = r["total"]
            result["offset"] = offset
            result["limit"] = limit
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteVenvHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            venv_id = self.get_argument("venv_id", "")
            if venv_id:
                success = VenvManager.instance().delete(venv_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class UpdateVenvHandler(StreamBaseHandler):
    @auth_check
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            venv_id = self.get_form_argument("venv_id", "")
            name = self.get_form_argument("name", "")
            description = self.get_form_argument("description", "")
            LOG.debug("UpdateVenvHandler venv_id: %s, name: %s, description: %s", venv_id, name, description)
            if venv_id and VenvManager.instance().info(venv_id):
                success = VenvManager.instance().update(venv_id, name, description, self.file_path.decode("utf-8"))
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        LOG.debug("update venv package use: %ss", time.time() - self.start)
        self.write(result)
        self.finish()


class InfoVenvHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            venv_id = self.get_argument("venv_id", "")
            if venv_id:
                venv_info = VenvManager.instance().info(venv_id)
                if venv_info:
                    result["venv_info"] = venv_info
                elif venv_info is None:
                    Errors.set_result_error("VenvNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DownloadVenvHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            venv_id = self.get_argument("venv_id", "")
            sha1 = self.get_argument("sha1", "")
            if venv_id:
                venv_info = VenvManager.instance().info(venv_id)
                if venv_info:
                    if sha1 == "":
                        sha1 = venv_info["sha1"]
                    f = VenvManager.instance().open(venv_id, sha1)
                    if f:
                        self.set_header('Content-Type', 'application/octet-stream')
                        if "venv_store" in CONFIG:
                            if "tar.gz" in CONFIG["venv_store"]:
                                self.set_header('Content-Disposition', 'attachment; filename=%s.tar.gz' % venv_id)
                            elif "zip" in CONFIG["venv_store"]:
                                self.set_header('Content-Disposition', 'attachment; filename=%s.zip' % venv_id)
                            else:
                                self.set_header('Content-Disposition', 'attachment; filename=%s.tar.gz' % venv_id)
                        else:
                            self.set_header('Content-Disposition', 'attachment; filename=%s.tar.gz' % venv_id)
                        buf_size = 1024 * 1024
                        while True:
                            data = f.read(buf_size)
                            if not data:
                                break
                            self.write(data)
                            self.flush()
                            yield gen.moment
                        f.close()
                        self.finish()
                        return
                    else:
                        Errors.set_result_error("OperationFailed", result)
                elif venv_info is None:
                    Errors.set_result_error("VenvNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.set_status(400)
        self.write(result)
        self.finish()


class VenvHistoryListHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK, "venv_histories": [], "total": 0}
        try:
            venv_id = self.get_argument("venv_id", "")
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            if venv_id:
                venv_info = VenvManager.instance().info(venv_id)
                if venv_info:
                    venv_histories = VenvManager.instance().list_history(offset, limit, {"venv_id": venv_id})
                    if venv_histories:
                        result["venv_histories"] = venv_histories["histories"]
                        result["total"] = venv_histories["total"]
                    result["offset"] = offset
                    result["limit"] = limit
                elif venv_info is None:
                    Errors.set_result_error("VenvNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class VenvHistoryInfoHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            venv_id = self.get_argument("venv_id", "")
            history_id = int(self.get_argument("history_id", "-1"))
            if history_id != -1:
                venv_history = VenvManager.instance().info_history(history_id, venv_id)
                if venv_history:
                    result["history_info"] = venv_history
                elif venv_history is None:
                    Errors.set_result_error("VenvHistoryNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class VenvHistoryActivateHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            history_id = self.get_json_argument("history_id", "")
            venv_id = self.get_json_argument("venv_id", "")
            if history_id and venv_id:
                success = VenvManager.instance().activate_history(history_id, venv_id = venv_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class VenvHistoryDeleteHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            venv_id = self.get_argument("venv_id", "")
            history_id = int(self.get_argument("history_id", "-1"))
            if history_id != -1 and venv_id:
                success = VenvManager.instance().delete_history(history_id, venv_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


# -*- coding: utf-8 -*-

import json
import time
import urllib
import logging

from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, BaseSocketHandler, auth_check
from litepipeline.manager.models.schedules import Schedules
from litepipeline.manager.models.workflows import Workflows
from litepipeline.manager.utils.app_manager import AppManager
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, Errors, Stage, splitall, JSONLoadError
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class CreateScheduleHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            schedule_name = self.get_json_argument("schedule_name", "")
            source = self.get_json_argument("source", "")
            source_id = self.get_json_argument("source_id", "")
            input_data = self.get_json_argument("input_data", {})
            minute = int(self.get_json_argument("minute", -1))             # [0, 59]
            hour = int(self.get_json_argument("hour", -1))                 # [0, 23]
            day_of_month = int(self.get_json_argument("day_of_month", -1)) # [1, 31]
            day_of_week = int(self.get_json_argument("day_of_week", -1))   # [1, 7] (Sunday = 7)
            enable = True if self.get_json_argument("enable", False) else False
            if (
                    (
                        (source == Schedules.application and AppManager.instance().info(source_id)) or 
                        (source == Schedules.workflow and Workflows.instance().get(source_id))
                    ) and
                    schedule_name and
                    (minute == -1 or (minute >= 0 and minute <= 59)) and
                    (hour == -1 or (hour >= 0 and hour <= 23)) and
                    (day_of_month == -1 or (day_of_month >= 1 and day_of_month <= 31)) and
                    (day_of_week == -1 or (day_of_week >= 1 and day_of_week <= 7))
                ):
                if not isinstance(input_data, dict):
                    raise JSONLoadError("input_data must be dict type")
                schedule_id = Schedules.instance().add(
                    schedule_name,
                    source,
                    source_id,
                    minute = minute,
                    hour = hour,
                    day_of_month = day_of_month,
                    day_of_week = day_of_week,
                    enable = enable,
                    input_data = input_data
                )
                if schedule_id is not False:
                    result["schedule_id"] = schedule_id
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("CreateScheduleHandler, schedule_name: %s, source: %s, source_id: %s", schedule_name, source, source_id)
        except JSONLoadError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class UpdateScheduleHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            schedule_id = self.get_json_argument("schedule_id", "")
            result["schedule_id"] = schedule_id
            data = self.get_json_exists_arguments(
                [
                    "schedule_name",
                    "source",
                    "source_id",
                    "input_data",
                    "minute",        # [0, 59]
                    "hour",          # [0, 23]
                    "day_of_month",  # [1, 31]
                    "day_of_week",   # [1, 7] (Sunday = 7)
                    "enable",
                ]
            )
            if (
                    data and
                    (schedule_id and Schedules.instance().get(schedule_id)) and
                    (("schedule_name" in data and data["schedule_name"] != "") or "schedule_name" not in data) and
                    (
                        "source_id" not in data or
                        ("source_id" in data and data["source"] == Schedules.application and AppManager.instance().info(data["source_id"])) or
                        ("source_id" in data and data["source"] == Schedules.workflow and Workflows.instance().get(data["source_id"]))
                    ) and
                    (("minute" in data and (data["minute"] == -1 or (data["minute"] >= 0 and data["minute"] <= 59))) or "minute" not in data) and
                    (("hour" in data and (data["hour"] == -1 or (data["hour"] >= 0 and data["hour"] <= 23))) or "hour" not in data) and
                    (("day_of_month" in data and (data["day_of_month"] == -1 or (data["day_of_month"] >= 1 and data["day_of_month"] <= 31))) or "day_of_month" not in data) and
                    (("day_of_week" in data and (data["day_of_week"] == -1 or (data["day_of_week"] >= 1 and data["day_of_week"] <= 7))) or "day_of_week" not in data) and
                    (("enable" in data and data["enable"] in [True, False]) or "enable" not in data)
                ):
                if "input_data" in data and not isinstance(data["input_data"], dict):
                    raise JSONLoadError("input_data must be dict type")
                if "app_id" in data:
                    data["application_id"] = data["app_id"]
                    del data["app_id"]
                success = Schedules.instance().update(schedule_id, data)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("UpdateScheduleHandler, schedule_id: %s, data: %s", schedule_id, data)
        except JSONLoadError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteScheduleHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            schedule_id = self.get_argument("schedule_id", "")
            if schedule_id:
                success = Schedules.instance().delete(schedule_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class InfoScheduleHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            schedule_id = self.get_argument("schedule_id", "")
            if schedule_id:
                schedule_info = Schedules.instance().get(schedule_id)
                if schedule_info:
                    result["schedule_info"] = schedule_info
                    if schedule_id in Schedules.instance().cache:
                        result["schedule_cache_info"] = Schedules.instance().cache[schedule_id]
                elif schedule_info is None:
                    Errors.set_result_error("ScheduleNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ListScheduleHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            filters = {}
            schedule_id = self.get_argument("schedule_id", "")
            if schedule_id:
                filters["schedule_id"] = schedule_id
            source_id = self.get_argument("source_id", "")
            if source_id:
                filters["source_id"] = source_id
            name = self.get_argument("name", "")
            if name:
                filters["name"] = name
            enable = self.get_argument("enable", "").lower()
            if enable == "true":
                enable = True
            elif enable == "false":
                enable = False
            else:
                enable = ""
            if enable is not "":
                filters["enable"] = enable
            LOG.debug("ListScheduleHandler offset: %s, limit: %s, filters: %s", offset, limit, filters)
            r = Schedules.instance().list(offset = offset, limit = limit, filters = filters)
            result["schedules"] = r["schedules"]
            result["total"] = r["total"]
            result["offset"] = offset
            result["limit"] = limit
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()



# -*- coding: utf-8 -*-

import json
import time
import logging

from tornado import web
from tornado import gen

from litepipeline.viewer.handlers.base import BaseHandler, BaseSocketHandler
from litepipeline.viewer.utils.common import encode_token
from litepipeline.viewer.config import CONFIG

LOG = logging.getLogger("__name__")


class ScheduleHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        self.render(
            "schedule/schedule.html",
            current_nav = "schedule",
            manager_host = "%s:%s" % (
                self.get_manager_http_host(),
                CONFIG["manager_http_port"]),
            user = CONFIG["manager_user"],
            token = encode_token(CONFIG["manager_user"], CONFIG["manager_password"])
        )



# -*- coding: utf-8 -*-

import os
import io
import re
import time
import logging
import hashlib
from base64 import b64encode, b64decode

from tornado import web
from tornado import locale
from tornado import websocket
from tea_encrypt import EncryptStr, DecryptStr

from litepipeline.manager.config import CONFIG
from litepipeline.manager.utils.common import bytes_md5sum, AuthError, Errors

LOG = logging.getLogger(__name__)


def auth_check(method):
    def nope(ref):
        pass

    def inner(ref):
        result = {"result": Errors.OK}
        user, password = ref.auth()
        if user and password:
            ref.user = user
            ref.password = password
            return method(ref)
        else:
            LOG.error("permission dined")
            Errors.set_result_error("AuthError", result)
            ref.write(result)
            ref.finish()
            return nope(ref)
    return inner


class BaseHandler(web.RequestHandler):
    # def get_current_user(self):
    #     return self.get_secure_cookie("user", max_age_days = 1)

    # def get_current_user_key(self):
    #     return self.get_secure_cookie("user_key", max_age_days = 1)

    # def get_user_locale_value(self):
    #     user_locale = self.get_secure_cookie("user_locale", max_age_days = 1)
    #     if user_locale:
    #         return user_locale
    #     return None

    def get_user(self, user):
        for user_info in CONFIG["users"]:
            u, password = user_info["name"], user_info["password"]
            if u == user:
                return bytes_md5sum(password.encode("utf-8"))
        return None

    def decode_token(self, token, password):
        return DecryptStr(b64decode(token.encode("utf-8")), password)

    def auth(self):
        result = None, None
        headers = self.request.headers
        user = None
        token = None
        try:
            if "user" in headers and "token" in headers:
                user = self.request.headers["user"]
                token = self.request.headers["token"]
                password = self.get_user(user)
                if password:
                    content = self.decode_token(token, password)
                    LOG.info("content: %s", content)
                    if content == user.encode("utf-8"):
                        result = user, password
            LOG.info("user: %s, token: %s", user, token)
        except Exception as e:
            LOG.exception(e)
        return result

    def set_default_headers(self):
        self.set_header("Content-Type", 'application/json')
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Expose-Headers", "Content-Disposition")
        self.set_header("Access-Control-Allow-Headers", "Content-Type, Access-Control-Allow-Headers, X-Requested-With, user, token, Content-Disposition")
        self.set_header('Access-Control-Allow-Methods', 'GET, POST, PUT, DELETE, OPTIONS')

    def options(self):
        self.set_status(204)
        self.finish()

    def get_user_locale(self):
        pass
        # user_locale = self.get_secure_cookie("user_locale", max_age_days = 1)
        # if user_locale:
        #     return tornado.locale.get(user_locale)
        # return None
    # def __init__(self, application, request, **kwargs):
    #     super(BaseHandler, self).__init__(application, request, kwargs)
    #     self.locale_code = "en_US"
    # pass

    def get_json_argument(self, key, default_value):
        result = default_value
        if key in self.json_data:
            result = self.json_data[key]
        return result

    def get_json_exists_arguments(self, keys):
        result = {}
        for key in keys:
            if key in self.json_data:
                result[key] = self.json_data[key]
        return result


class BaseSocketHandler(websocket.WebSocketHandler):
    def get_current_user(self):
        pass
        # return self.get_secure_cookie("user", max_age_days = 1)

    # def get_current_user_key(self):
    #     return self.get_secure_cookie("user_key", max_age_days = 1)

    # def get_user_locale(self):
    #     user_locale = self.get_secure_cookie("user_locale", max_age_days = 1)
    #     if user_locale:
    #         return tornado.locale.get(user_locale)
    #     return None


@web.stream_request_body
class StreamBaseHandler(BaseHandler):
    PARSE_READY = 0
    PARSE_FILE_PENDING = 1

    def check_xsrf_cookie(self): # ignore xsrf check
        pass

    def prepare(self):
        self.mimetype = self.request.headers.get("Content-Type")
        self.boundary = ""
        if self.mimetype:
            self.boundary = "--%s" % (self.mimetype[self.mimetype.find("boundary")+9:])
        self.boundary = self.boundary.encode("utf-8")
        self.state = StreamBaseHandler.PARSE_READY
        self.output = None
        self.find_filename = re.compile(b'filename="(.*)"')
        self.find_mimetype = re.compile(b'Content-Type: (.*)')
        self.find_field = re.compile(b'name="(.*)"')
        self.start = time.time()
        self.file_name = ""
        self.file_path = b""
        self.form_arguments = {}

    def data_received(self, data):
        LOG.debug("chunk size: %s", len(data))
        try:
            buff = data.split(self.boundary)
            for index, part in enumerate(buff):
                if part:
                    if part == b"--\r\n":
                        break
                    if self.state == StreamBaseHandler.PARSE_FILE_PENDING:
                        if len(buff) > 1:
                            self.output.write(part[:-2])
                            self.output.close()
                            self.state = StreamBaseHandler.PARSE_READY
                            continue
                        else:
                            self.output.write(part)
                            self.output.flush()
                            continue

                    elif self.state == StreamBaseHandler.PARSE_READY:
                        stream = io.BytesIO(part)
                        stream.readline()
                        form_data_type_line = stream.readline()
                        if form_data_type_line.find(b"filename") > -1:
                            filename = re.search(self.find_filename, form_data_type_line).groups()[0]
                            self.file_name = os.path.split(filename)[-1]
                            self.file_path = os.path.join(CONFIG["data_path"].encode("utf-8"), b"tmp", self.file_name)
                            if self.file_name:
                                self.output = open(self.file_path, "wb")
                                content_type_line = stream.readline()
                                mimetype = re.search(self.find_mimetype, content_type_line).groups()[0]
                                LOG.debug("%s with %s" % (filename, mimetype.strip()))
                                stream.readline()
                                body = stream.read()
                                if len(buff) > index + 1:
                                    self.output.write(body[:-2])
                                    self.output.flush()
                                    self.state = StreamBaseHandler.PARSE_READY
                                else:
                                    self.output.write(body)
                                    self.output.flush()
                                    self.state = StreamBaseHandler.PARSE_FILE_PENDING
                        else:
                            stream.readline()
                            form_name = re.search(self.find_field, form_data_type_line).groups()[0]
                            form_value = stream.readline()
                            if form_name:
                                self.form_arguments[form_name.strip().decode("utf-8")] = form_value.strip().decode("utf-8")
                            self.state = StreamBaseHandler.PARSE_READY
                            LOG.debug("%s = %s" % (form_name.strip(), form_value.strip()))
        except Exception as e:
            LOG.exception(e)

    def get_form_argument(self, key, default_value):
        result = default_value
        try:
            if key in self.form_arguments:
                result = self.form_arguments[key]
        except Exception as e:
            LOG.exception(e)
        return result



# -*- coding: utf-8 -*-

import os
import time
import json
import logging
import shutil
import tarfile

import requests
from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, StreamBaseHandler, auth_check
from litepipeline.manager.models.workflows import Workflows
from litepipeline.manager.utils.listener import Connection
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, Errors, splitall
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class CreateWorkflowHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            name = self.get_json_argument("name", "")
            configuration = self.get_json_argument("configuration", {})
            description = self.get_json_argument("description", "")
            enable = self.get_json_argument("enable", False)
            if name:
                workflow_id = Workflows.instance().add(name, configuration, description = description, enable = enable)
                result["workflow_id"] = workflow_id
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ListWorkflowHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            filters = {}
            workflow_id = self.get_argument("id", "")
            if workflow_id:
                filters["id"] = workflow_id
            name = self.get_argument("name", "")
            if name:
                filters["name"] = name
            LOG.debug("ListWorkflowHandler offset: %s, limit: %s, filters: %s", offset, limit, filters)
            r = Workflows.instance().list(offset = offset, limit = limit, filters = filters)
            result["workflows"] = r["workflows"]
            result["total"] = r["total"]
            result["offset"] = offset
            result["limit"] = limit
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteWorkflowHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            workflow_id = self.get_argument("workflow_id", "")
            if workflow_id:
                success = Workflows.instance().delete(workflow_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class UpdateWorkflowHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            workflow_id = self.get_json_argument("workflow_id", "")
            name = self.get_json_argument("name", "")
            configuration = self.get_json_argument("configuration", {})
            description = self.get_json_argument("description", "")
            enable = self.get_json_argument("enable", None)
            LOG.debug("UpdateWorkflowHandler workflow_id: %s, name: %s, description: %s", workflow_id, name, description)
            if workflow_id and Workflows.instance().get(workflow_id):
                result["workflow_id"] = workflow_id
                data = {}
                if name:
                    data["name"] = name
                if configuration:
                    data["configuration"] = configuration
                if description:
                    data["description"] = description
                if enable is not None:
                    data["enable"] = enable
                if data:
                    success = Workflows.instance().update(workflow_id, data)
                    if not success:
                        Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class InfoWorkflowHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            workflow_id = self.get_argument("workflow_id", "")
            if workflow_id:
                workflow_info = Workflows.instance().get(workflow_id)
                if workflow_info:
                    result["info"] = workflow_info
                elif workflow_info is None:
                    Errors.set_result_error("AppNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()
